@@ -177,10 +177,6 @@ class Baseset:
     def gen_batch(self, batch_size, shuffle=False, n_iters=None, n_epochs=None, drop_last=False,
                   notifier=False, *args, **kwargs):
         """ Generate batches """
-        if n_epochs is None and n_iters is None:
-            warnings.warn('Batch generation will never stop as ' \
-                          'n_epochs=None and n_iters=None', RuntimeWarning)
-
         iter_params = kwargs.pop('iter_params', None)
         for ix_batch in self.index.gen_batch(batch_size, shuffle, n_iters, n_epochs, drop_last,
                                              notifier, iter_params):
