@@ -145,13 +145,16 @@ class DecoderModule(nn.ModuleDict):
                     layer_desc = 'upsample-{}'.format(i)
 
                 elif letter in ['c']:
-                    if self.skip and (i < len(inputs) - 2):
-                        args = {'factor': factor[i],
-                                **kwargs, **combine_args, **unpack_args(combine_args, i, num_stages)}
+                    if self.skip:
+                        if i < len(inputs) - 2:
+                            args = {'factor': factor[i],
+                                    **kwargs, **combine_args, **unpack_args(combine_args, i, num_stages)}
 
-                        layer = Combine(inputs=[x, inputs[-i - 3]], **args)
-                        x = layer([x, inputs[-i - 3]])
-                        layer_desc = 'combine-{}'.format(i)
+                            layer = Combine(inputs=[x, inputs[-i - 3]], **args)
+                            x = layer([x, inputs[-i - 3]])
+                            layer_desc = 'combine-{}'.format(i)
+                    else:
+                        continue
                 else:
                     raise ValueError('Unknown letter in order {}, use one of ("b", "u", "c")'.format(letter))
 
@@ -203,7 +206,7 @@ class Encoder(TorchModel):
     @classmethod
     def body(cls, inputs, return_all=False, **kwargs):
         """ Make a sequential list of encoder modules. """
-        kwargs = cls.get_defaults('body', kwargs)
+        kwargs = cls.get_block_defaults('body', kwargs)
         encoder = kwargs.pop('encoder')
         layers = [('encoder', EncoderModule(inputs=inputs, return_all=return_all, **{**kwargs, **encoder}))]
         return nn.Sequential(OrderedDict(layers))
@@ -271,17 +274,17 @@ class Decoder(TorchModel):
     @classmethod
     def body(cls, inputs, **kwargs):
         """ Make a sequential list of decoder modules. """
-        kwargs = cls.get_defaults('body', kwargs)
+        kwargs = cls.get_block_defaults('body', kwargs)
         decoder = kwargs.pop('decoder')
         layers = [('decoder', DecoderModule(inputs=inputs, **{**kwargs, **decoder}))]
         return nn.Sequential(OrderedDict(layers))
 
     @classmethod
-    def head(cls, inputs, target_shape, classes, **kwargs):
+    def head(cls, inputs, classes=None, **kwargs):
         """ Make network's head. If needed, apply 1x1 convolution to obtain correct output shape. """
-        kwargs = cls.get_defaults('head', kwargs)
+        kwargs = cls.get_block_defaults('head', kwargs)
         layers = []
-        layer = super().head(inputs, target_shape, classes, **kwargs)
+        layer = super().head(inputs, classes=classes, **kwargs)
         if layer is not None:
             inputs = layer(inputs)
             layers.append(layer)
@@ -430,7 +433,7 @@ class EncoderDecoder(Decoder):
     @classmethod
     def body(cls, inputs, **kwargs):
         """ Sequence of encoder, embedding and decoder. """
-        kwargs = cls.get_defaults('body', kwargs)
+        kwargs = cls.get_block_defaults('body', kwargs)
         encoder = kwargs.pop('encoder')
         embedding = kwargs.pop('embedding')
         decoder = kwargs.pop('decoder')
