@@ -1,4 +1,6 @@
 """ Contains the base class for open datasets """
+import numpy as np
+
 from .. import Dataset, DatasetIndex
 from .. import ImagesBatch
 
@@ -39,6 +41,13 @@ class Openset(Dataset):
         test_index = DatasetIndex(list(range(train_len, total_len)))
         return index, train_index, test_index
 
+    def create_array(self, images):
+        """ Create numpy array of objects. """
+        array = np.empty(len(images), dtype=object)
+        for i, image in enumerate(images):
+            array[i] = image
+        return array
+
 
 class ImagesOpenset(Openset):
     """ The base class for open datasets with images """
