@@ -6,15 +6,8 @@ from functools import partial
 import threading
 import asyncio
 import logging
-from pstats import Stats
-from cProfile import Profile
 import warnings
 
-try:
-    import pandas as pd
-except ImportError:
-    from . import _fake as pd
-
 from .base import Baseset
 from .config import Config
 from .batch import Batch
@@ -31,6 +24,7 @@ from ._const import *       # pylint:disable=wildcard-import
 from .utils import save_data_to
 from .utils_random import make_rng
 from .pipeline_executor import PipelineExecutor
+from .profiler import PipelineProfiler
 
 
 METRICS = dict(
@@ -59,7 +53,7 @@ def hashable(x):
 
 class Pipeline:
     """ Pipeline """
-    def __init__(self, dataset=None, config=None, pipeline=None, actions=None, proba=None, repeat=None):
+    def __init__(self, dataset=None, config=None, pipeline=None, actions=None, strict=False, proba=None, repeat=None):
         # pylint: disable=protected-access
 
         if pipeline is None:
@@ -68,7 +62,8 @@ class Pipeline:
             self._actions = actions or []
             self._lazy_run = None
             self.models = ModelDirectory()
-            self.variables = VariableDirectory()
+            self.variables = VariableDirectory(strict)
+            self.strict = strict
             self.before = OncePipeline(self)
             self.after = OncePipeline(self)
             self._namespaces = []
@@ -87,6 +82,7 @@ class Pipeline:
                         self._actions[-1]['repeat'] = mult_option(repeat, self.get_last_action_repeat())
             self._lazy_run = pipeline._lazy_run
             self.variables = pipeline.variables.copy()
+            self.strict = pipeline.strict
             self.models = pipeline.models.copy()
             self._namespaces = pipeline._namespaces
             self.before = pipeline.before.copy()
@@ -106,20 +102,16 @@ class Pipeline:
         self.random_seed = None
 
         self._profiler = None
-        self.profile_info = None
-        self._profile_info_lock = threading.Lock()
 
     def __getstate__(self):
         state = self.__dict__.copy()
         state['random'] = getattr(self._local, 'random', None)
         state.pop('_local')
-        state.pop('_profile_info_lock')
         state['_profiler'] = None
         state['_batch_generator'] = None
         return state
 
     def __setstate__(self, state):
-        self._profile_info_lock = threading.Lock()
         self._local = threading.local()
         for k, v in state.items():
             setattr(self, k, v)
@@ -338,6 +330,47 @@ class Pipeline:
         """ Add a nested pipeline to the log of future actions """
         self._actions.append({'name': PIPELINE_ID, 'pipeline': pipeline, 'proba': proba, 'repeat': repeat})
 
+
+    LINE_LENGTH = 80
+    ITEM_LENGTH = 40
+    PREFIX = '\033[1m\033[4m'
+
+    def __str__(self):
+        """ Textual representation of a pipeline: list all actions and their keyword parameters. """
+        # TODO: with the help of `inspect` and `get_method` method, can match the `args` to their actual names
+        msg = []
+        for i, action in enumerate(self._actions):
+            name = self.get_action_name(action)
+            indent = (len(name) + len(str(i)) + 3) * ' '
+
+            # Combine kwargs passed directly to actions and ones from pipeline methods
+            kwargs = {**action['kwargs'], **action}
+            for key in ['name', 'args', 'kwargs', 'proba', 'repeat']:
+                kwargs.pop(key)
+
+            # Make string with parameters
+            lines_kwargs = []
+            str_kwargs = ''
+            for key, value in kwargs.items():
+                # Control the length of one parameter value
+                value_ = repr(value)
+                value_ = value_ if len(value_) < self.ITEM_LENGTH else '[...]'
+                str_kwargs += f'{key}={value_}, '
+
+                # Control the length of current line
+                if len(str_kwargs) + len(indent) > self.LINE_LENGTH:
+                    lines_kwargs.append(str_kwargs[:-2])
+                    str_kwargs = ''
+            if str_kwargs:
+                lines_kwargs.append(str_kwargs[:-2])
+            str_kwargs = f'\n{indent}'.join(lines_kwargs)
+
+            action_msg = f'#{i} {self.PREFIX}{name}\033[0m({str_kwargs})'
+            msg.append(action_msg)
+
+        return '\n'.join(msg)
+
+
     @property
     def index(self):
         """ Return index of the source dataset """
@@ -597,13 +630,9 @@ class Pipeline:
         """
         V(name, mode=mode).set(value, batch=batch, pipeline=self)
 
-    def assign_variable(self, name, value, **kwargs):
+    def assign_variable(self, name, value):
         """ Assign a value to a variable """
-        _ = kwargs
-        if not self.has_variable(name):
-            logging.warning("Pipeline variable '%s' has not been initialized", name)
-            self.init_variable(name)
-        self.variables.set(name, value)
+        self.variables.set(name, value, pipeline=self)
 
     def delete_variable(self, name):
         """ Delete a variable
@@ -839,99 +868,6 @@ class Pipeline:
                 batch = self._exec_all_actions(batch, action['pipeline']._actions)  # pylint: disable=protected-access
         return batch
 
-    def _add_profile_info(self, batch, action, exec_time, **kwargs):
-        name = self.get_action_name(action, add_index=True)
-        iter_no = batch.iteration
-
-        if isinstance(self._profiler, Profile):
-            stats = Stats(self._profiler)
-            self._profiler.clear()
-
-            indices, values = [], []
-            for key, value in stats.stats.items():
-                for k, v in value[4].items():
-                    # action name, method_name, file_name, line_no, callee
-                    indices.append((name, '{}::{}::{}::{}'.format(key[2], *k)))
-                    row_dict = {
-                        'iter': iter_no, 'total_time': exec_time, 'pipeline_time': stats.total_tt, # base stats
-                        'ncalls': v[0], 'tottime': v[2], 'cumtime': v[3], # detailed stats
-                        'batch_id': id(batch), **kwargs
-                    }
-                    values.append(row_dict)
-        else:
-            indices = [(name, '')]
-            values = [{'iter': iter_no, 'total_time': exec_time, 'pipeline_time': exec_time, 'batch_id': id(batch),
-                       **kwargs}]
-
-        multiindex = pd.MultiIndex.from_tuples(indices, names=['action', 'id'])
-        df = pd.DataFrame(values, index=multiindex)
-
-        with self._profile_info_lock:
-            if self.profile_info is None:
-                self.profile_info = df
-            else:
-                self.profile_info = self.profile_info.append(df)
-
-    def show_profile_info(self, per_iter=False, detailed=False,
-                          groupby=None, columns=None, sortby=None, limit=10):
-        """ Show stored profiling information with varying levels of details.
-
-        Parameters
-        ----------
-        per_iter : bool
-            Whether to make an aggregation over iters or not.
-        detailed : bool
-            Whether to use information from :class:`cProfiler` or not.
-        groupby : str or sequence of str
-            Used only when `per_iter` is True, directly passed to pandas.
-        columns : sequence of str
-            Columns to show in resultining dataframe.
-        sortby : str or tuple of str
-            Column id to sort on. Note that if data is aggregated over iters (`per_iter` is False),
-            then it must be a full identificator of a column.
-        limit : int
-            Limits the length of resulting dataframe.
-        parse : bool
-            Allows to re-create underlying dataframe from scratches.
-        """
-        if self.profile_info is None:
-            warnings.warn("Profiling has not been enabled.")
-            return None
-
-        detailed = False if not isinstance(self._profiler, Profile) else detailed
-
-        if per_iter is False and detailed is False:
-            columns = columns or ['total_time', 'pipeline_time']
-            sortby = sortby or ('total_time', 'sum')
-            aggs = {key: ['sum', 'mean', 'max'] for key in columns}
-            result = (self.profile_info.groupby(['action', 'iter'])[columns].mean().groupby('action').agg(aggs)
-                      .sort_values(sortby, ascending=False))
-
-        elif per_iter is False and detailed is True:
-            columns = columns or ['ncalls', 'tottime', 'cumtime']
-            sortby = sortby or ('tottime', 'sum')
-            aggs = {key: ['sum', 'mean', 'max'] for key in columns}
-            result = (self.profile_info.reset_index().groupby(['action', 'id']).agg(aggs)
-                      .sort_values(['action', sortby], ascending=[True, False])
-                      .groupby(level=0).apply(lambda df: df[:limit]).droplevel(0))
-
-        elif per_iter is True and detailed is False:
-            groupby = groupby or ['iter', 'action']
-            columns = columns or ['action', 'total_time', 'pipeline_time', 'batch_id']
-            sortby = sortby or 'total_time'
-            result = (self.profile_info.reset_index().groupby(groupby)[columns].mean()
-                      .sort_values(['iter', sortby], ascending=[True, False]))
-
-        elif per_iter is True and detailed is True:
-            groupby = groupby or ['iter', 'action', 'id']
-            columns = columns or ['ncalls', 'tottime', 'cumtime']
-            sortby = sortby or 'tottime'
-            result = (self.profile_info.reset_index().set_index(groupby)[columns]
-                      .sort_values(['iter', 'action', sortby], ascending=[True, True, False])
-                      .groupby(level=[0, 1]).apply(lambda df: df[:limit]).droplevel([0, 1]))
-        return result
-
-
     def _exec_all_actions(self, batch, actions=None, iteration=None):
         join_batches = None
         actions = actions or self._actions
@@ -941,9 +877,7 @@ class Pipeline:
 
         for action in actions:
             if self._profiler:
-                start_time = time.time()
-                if isinstance(self._profiler, Profile):
-                    self._profiler.enable()
+                self._profiler.enable()
 
             if action.get('#dont_run', False):
                 pass
@@ -974,17 +908,21 @@ class Pipeline:
                     action['args'] = join_batches + action['args']
                     join_batches = None
 
-                batch, eval_time = self._exec_one_action(batch, action, iteration=iteration)
+                batch, action_time = self._exec_one_action(batch, action, iteration=iteration)
 
             if self._profiler:
-                if isinstance(self._profiler, Profile):
-                    self._profiler.disable()
-                exec_time = time.time() - start_time
-                self._add_profile_info(batch, action, start_time=start_time, exec_time=exec_time,
-                                       eval_time=eval_time)
+                name = self.get_action_name(action, add_index=True)
+                self._profiler.disable(batch.iteration, name, batch_id=id(batch), action_time=action_time)
 
         return batch
 
+    def show_profile_info(self, **kwargs):
+        return self._profiler.show_profile_info(**kwargs)
+
+    @property
+    def profile_info(self):
+        return self._profiler.profile_info
+
     def _needs_exec(self, batch, action):
         if action['proba'] is None:
             return True
@@ -1036,8 +974,17 @@ class Pipeline:
         """ A shorter alias for get_model_by_name() """
         return self.get_model_by_name(name, batch=batch)
 
-    def init_model(self, name, model_class=None, mode='dynamic', config=None):
-        """ Initialize a static or dynamic model
+    @property
+    def model(self):
+        """ An alias for a present model, if pipeline has only one model initialized. """
+        n_models = len(self.models.models)
+        if n_models == 1:
+            return list(self.models.models.values())[0]
+        raise TypeError('`model` property should be used only for pipelines with exactly 1 initialized model, '
+                        f'but pipeline has {n_models} instead!')
+
+    def init_model(self, name, model_class=None, mode='dynamic', config=None, source=None):
+        """ Initialize a static or dynamic model by building or importing it
 
         Parameters
         ----------
@@ -1055,23 +1002,34 @@ class Pipeline:
         config : dict or Config
             model configurations parameters, where each key and value could be named expressions.
 
+        source
+            a model or a pipeline to import from
+
         Examples
         --------
-        >>> pipeline.init_model('my-model', MyModel, 'static')
+        Build a model::
 
-        >>> pipeline
+            pipeline.init_model('my-model', MyModel, 'static')
+
+        Import a model::
+
+            pipeline.init_model('my-model', source=train_pipeline)
+
+        Build a model with a config::
+
+            pipeline
               .init_variable('images_shape', [256, 256])
               .init_model('my_model', MyModel, 'static', config={'input_shape': V('images_shape')})
 
-        >>> pipeline
+            pipeline
               .init_variable('shape_name', 'images_shape')
-              .init_model('my_model', C('model'), 'dynamic', config={V('shape_name)': B('images_shape')})
+              .init_model('my_model', C('model'), 'dynamic', config={V('shape_name'): B('images_shape')})
 
         """
-        self.before.init_model(name, model_class, mode=mode, config=config)
+        self.before.init_model(name, model_class, mode=mode, config=config, source=source)
         return self
 
-    def import_model(self, name, model):
+    def import_model(self, name, source):
         """ Import a model from another pipeline
 
         Parameters
@@ -1079,7 +1037,7 @@ class Pipeline:
         name : str
             a name with which the model is stored in this pipeline
 
-        model : model or pipeline
+        source
             a model or a pipeline to import from
 
         Examples
@@ -1096,7 +1054,7 @@ class Pipeline:
 
             pipeline.import_model('my-model', train_pipeline)
         """
-        return self._add_action(IMPORT_MODEL_ID, _args=dict(source=model, model_name=name))
+        return self._add_action(IMPORT_MODEL_ID, _args=dict(source=source, model_name=name))
 
     def _exec_import_model(self, batch, action):
         model_name = self._eval_expr(action['model_name'], batch=batch)
@@ -1182,11 +1140,13 @@ class Pipeline:
 
         >>> pipeline
             .init_variable('inferred_masks', default=[])
-            .predict_model('tf_unet', fetches='predictions', feed_dict={'x': B('images')},
+            .predict_model('my_model',
+                           B.images,
+                           fetches='predictions',
                            save_to=V('inferred_masks'))
 
-        Call a `tf_unet` model `train` method with `fetches` and `feed_dict` arguments:
-        ``predictions = tf_unet.train(fetches='predictions', feed_dict={'x': batch.images})``
+        Call a `my_model` model `train` method with images as positional and `fetches` as keyword arguments:
+        ``predictions = my_model.train(B.images, fetches='predictions')``
         Predictions for each batch will be stored in a pipeline variable `inferred_masks`.
         """
         return self._add_action(PREDICT_MODEL_ID, *args, _args=dict(model_name=name, save_to=save_to), **kwargs)
@@ -1442,7 +1402,7 @@ class Pipeline:
             whether to use profiler
 
         random
-            a random state (see :fun:`~.make_rng`).
+            a random state (see :func:`~.make_rng`).
             If not specified, RNG will be created with a random entropy.
 
         Examples
@@ -1486,10 +1446,10 @@ class Pipeline:
         self.random_seed = seed
 
         if profile == 2 or isinstance(profile, str) and 'detailed'.startswith(profile):
-            self._profiler = Profile()
-        elif profile is True or profile == 1:
-            self._profiler = True
-        else:
+            self._profiler = PipelineProfiler(detailed=True)
+        elif profile == 1 or profile is True:
+            self._profiler = PipelineProfiler(detailed=False)
+        else: # 0, False, None
             self._profiler = None
 
 
@@ -1599,7 +1559,7 @@ class Pipeline:
         target : 'threads' or 'mpc'
             batch parallelization engine used for prefetching (default='threads').
             'mpc' rarely works well due to complicated and slow python's inter-process communications.
-            Don't use pipeline variables and models in mpc-mode as each bach is being processed in
+            Don't use pipeline variables and models in mpc-mode as each batch is being processed in
             a separate copy of the pipeline.
 
         reset : list of str, str or bool
@@ -1635,6 +1595,8 @@ class Pipeline:
         args_value, kwargs_value = self._eval_run_args(args, kwargs)
 
         rebatch = len(self._actions) > 0 and self._actions[0]['name'] == REBATCH_ID
+        if rebatch:
+            _, self._actions[0] = self._eval_run_args([], self._actions[0])
 
         return PipelineExecutor(self).gen_batch(*args_value, dataset=self._dataset, rebatch=rebatch, **kwargs_value)
 
@@ -1645,22 +1607,25 @@ class Pipeline:
         batch_res = self.execute_for(batch)
         return batch_res
 
-    def next_batch(self, *args, **kwargs):
+    def next_batch(self, *args, n_epochs=None, **kwargs):
         """ Get the next batch and execute all lazy actions
 
+        Notes
+        -----
+        `n_epochs` is None by default to allow for infinite batch generation.
+
         See also
         --------
         :meth:`~.Pipeline.gen_batch`
         """
-        if len(args) == 0 and len(kwargs) == 0:
-            if self._lazy_run is None:
-                raise RuntimeError("next_batch without arguments requires a lazy run at the end of the pipeline")
+        if len(args) == 0 and len(kwargs) == 0 and self._lazy_run is not None:
             args, kwargs = self._lazy_run
-            batch = self.next_batch(*args, **kwargs)
         else:
-            if self._batch_generator is None:
-                self._batch_generator = self.gen_batch(*args, reset=None, **kwargs)
-            batch = next(self._batch_generator)
+            kwargs['n_epochs'] = n_epochs
+
+        if self._batch_generator is None:
+            self._batch_generator = self.gen_batch(*args, reset=None, **kwargs)
+        batch = next(self._batch_generator)
 
         return batch
 
@@ -1676,8 +1641,15 @@ class Pipeline:
             self._lazy_run = args, kwargs
             return self
 
+        if len(args) == 0 and len(kwargs) == 0 and self._lazy_run is not None:
+            args, kwargs = self._lazy_run
+
         args_value, kwargs_value = self._eval_run_args(args, kwargs)
 
+        if kwargs_value.get('n_epochs', None) is None and kwargs_value.get('n_iters', None) is None:
+            warnings.warn('Batch generation will never stop as ' \
+                          'n_epochs=None and n_iters=None', RuntimeWarning)
+
         return PipelineExecutor(self).run(*args_value, **kwargs_value)
 
     def run_now(self, *args, **kwargs):
